from __future__ import print_function
import pytest


from virgotools.frame_lib import (
    FrameFile, ChannelNotFound, expand_channel, expand_ffl, decode_sms,
    getChannel)
from virgotools import frame_lib
from virgotools.gpstime import *
from virgotools.pvt_utils import ascii
import numpy as np
import os
import mock

def test_expand_channel():

    with mock.patch('virgotools.frame_lib.DEFAULT_CHANNEL_PREFIX',b'V1:'):
        assert expand_channel('') == b'V1:'
        assert expand_channel('a') == b'V1:a'
        assert expand_channel('ab') == b'V1:ab'
        assert expand_channel('abc') == b'V1:abc'
        assert expand_channel('abcd') == b'V1:abcd'
    assert expand_channel('V1:chan') == b'V1:chan'
    assert expand_channel('H1:chan') == b'H1:chan'

    # change prefix
    frame_lib.DEFAULT_CHANNEL_PREFIX = b'K1:'
    assert expand_channel('chan') == b'K1:chan'
    # TODO test reading default_prefix from env using monkeypatch.setenv

def test_expand_ffl():
    # files ending in gwf and ffl should remain unmodified
    for fname in ['/path/to/frame_file.gwf', 'ffl_file.ffl']:
        assert expand_ffl(fname) == ascii(fname)
    with mock.patch('virgotools.frame_lib.DEFAULT_FFL_PATH', b'/virgoData/ffl'):
        assert expand_ffl('raw') == b'/virgoData/ffl/raw.ffl'
    # TODO mock FFLPATH environmental variable


def test_decode_sms():
    sms = b'a 1 b 2.3 units Volt c 4e56 d 0'
    assert list(decode_sms(sms)) == [
        (b'a', 1.0, b'Counts.'),
        (b'b', 2.3, b'Counts.'),
        (b'c', 4e+56, b'Volt'),
        (b'd', 0.0, b'Volt'),
    ]

@pytest.mark.parametrize('ffl_name', 'data/test.ffl'.split())
def test_ffl(ffl_name):
    print('\nReading from ffl file %s, this might take some time ...' % ffl_name)
    with FrameFile(ffl_name) as ffl:
        with ffl.first_frame() as frame:
            print('Gps of first frame of {}: {}'.format(
                frame_lib.expand_ffl(ffl_name), gps2str(frame.gps)))
            assert frame.gps == ffl.gps_start
            assert frame.gap == -1

        with next(ffl) as frame:
            print('Gps of next frame of {}: {}'.format(
                frame_lib.expand_ffl(ffl_name), gps2str(frame.gps)))
            assert frame.gap == 0

        with ffl.last_frame() as frame:
            print('Gps of last frame of {}:  {}'.format(
                frame_lib.expand_ffl(ffl_name), gps2str(frame.gps)))
            assert frame.gps + frame.dt == ffl.gps_end
            assert frame.gap == -1

            it = frame.iter_adc()
            adc = next(it)
            chan = adc.contents.name.data  # FIXME str() gives an error, might be a bug in ctypes wrapper
            print('Getting first channel:', chan)
            x = frame.get_adc_channel(chan)
            print('Got vector with shape', x.shape)

            for adc in it:  # iterate until end
                pass
            chan = adc.contents.name.data
            print('Last channel:', chan)

            with pytest.raises(ChannelNotFound):
                x = frame.SomeNonExistingChannel

        x = ffl.getChannel(chan, ffl.gps_end - 1, 10)
        print('Got channel %s with fsample = %g and shape %s' % (chan, x.fsample, x.data.shape))

        with pytest.raises(ChannelNotFound):
            x = ffl.getChannel('NonExistingChannel', ffl.gps_end - 1, 10)

        # ask for non-existing frame
        with pytest.raises(ValueError):
            frame = ffl.get_frame(1234)


# this requires data with a known gap
# see the data in mentioned here: https://github.com/gwpy/gwpy/pull/911#issuecomment-432748916
source = 'trend'
chan = 'V1:DAQ_FFL_RAW_LATENCY'
g0 = 1222585200  # exact 1800 sec frame boundary
span_good = g0, g0 + 1800 # no missing samples

# first and last data is ok, gap in the middle
span_across_gap = g0, g0 + 6 * 1800

# missing samples towards the end
span_last_samples_missing = g0, g0 + 2 * 1800

# first samples are missing, but there are good samples later in this frame
span_first_samples_missing = g0 + 4 * 1800, g0 + 6 * 1800

# channel is missing for the complete first frame, but data is present in
# later ones. I believed this failed, but seems to work anyhow.
span_start_complete_missing = g0 + 3 * 1800, g0 + 6 * 1800

# data completely missing, no hope of reading anything
span_completely_missing = g0 + 3 * 1800, g0 + 4 * 1800

fflfilename = frame_lib.DEFAULT_FFL_PATH + ascii("/" + source + ".ffl")
@pytest.mark.skipif(os.path.exists(fflfilename)==False,reason="The trend.ffl does not exist")
def test_missing_samples():
    x = getChannel(source, chan, *span_good)
    assert np.isfinite(x.data).all()
    x = getChannel(source, chan, *span_good, mask=True)
    assert x.missing is None

    for span in (span_across_gap, span_last_samples_missing,
                 span_first_samples_missing, span_start_complete_missing):
        x = getChannel(source, chan, *span, mask=True)
        assert x.missing is not None, 'expecting a mask vector'
        miss = x.missing == 0
        print('Found %i missing samples' % miss.sum())
        assert miss.any(), 'some samples in the mask should be True'
        y = getChannel(source, chan, *span)
        assert np.isnan(y.data[miss]).all(), 'all missing should be NaN'

    for span in (span_completely_missing,):
        with pytest.raises(ChannelNotFound):
            x = getChannel(source, chan, *span)
